/*
   ReactOS specific functions for UHCI module
   by Aleksey Bragin (aleksey@reactos.com)
   Some parts of code are inspired (or even just copied) from ReactOS Videoport driver
*/
#include <windows.h>
#include <initguid.h>
#include <ddk/usbdi.h>
#include <ddk/usbiodef.h>
#include <ddk/ntddk.h>
#define DEBUG
#include <debug.h>

// config and include core/hcd.h, for hc_device struct struct usb_interface *usb_ifnum_to_if(struct usb_device *dev, unsigned ifnum)

#include "../usb_wrapper.h"
#include "../core/hcd.h"

#include "../host/ohci_main.h"

// declare basic init funcs
void init_wrapper(struct pci_dev *probe_dev);
int uhci_hcd_init(void);
void uhci_hcd_cleanup(void);
int STDCALL usb_init(void);
void STDCALL usb_exit(void);

extern struct pci_driver uhci_pci_driver;
extern struct pci_device_id uhci_pci_ids[];
extern struct usb_driver hub_driver;

// dev->data = struct usb_hcd

extern NTSTATUS STDCALL UhciClose( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);
extern NTSTATUS STDCALL UhciCreate( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);
extern NTSTATUS STDCALL UsbDeviceControl(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);

extern NTSTATUS UsbDeviceControlPdo(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);
extern NTSTATUS UsbDeviceControlFdo( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);

// This should be removed, but for testing purposes it's here
//struct pci_dev *dev;
//struct pci_device_id *dev_id;

static ULONG DeviceNumber = 0;

#define USB_UHCI_TAG TAG('u','s','b','u')

NTSTATUS STDCALL 
AddDevice(PDRIVER_OBJECT DriverObject, PDEVICE_OBJECT pdo)
{
	PDEVICE_OBJECT fdo;
	NTSTATUS Status;
	WCHAR DeviceBuffer[20];
	WCHAR LinkDeviceBuffer[20];
	UNICODE_STRING DeviceName;
	UNICODE_STRING LinkDeviceName;
	POHCI_DRIVER_EXTENSION DriverExtension;
	POHCI_DEVICE_EXTENSION DeviceExtension;
//	ULONG Size;

	DPRINT1("UHCI: AddDevice called\n");

	if (pdo == NULL)
		DPRINT1("UHCI Pdo is Zero!\n");

	// Allocate driver extension now
	DriverExtension = IoGetDriverObjectExtension(DriverObject, DriverObject);
	if (DriverExtension == NULL)
	{
		Status = IoAllocateDriverObjectExtension(
					DriverObject,
					DriverObject,
					sizeof(OHCI_DRIVER_EXTENSION),
					(PVOID *)&DriverExtension);

		if (!NT_SUCCESS(Status))
		{
			DPRINT1("Allocating DriverObjectExtension failed.\n");
			return Status;
		}
	}
   
	// Create a unicode device name
//	DeviceNumber = 0; //TODO: Allocate new device number every time
	swprintf(DeviceBuffer, L"\\Device\\USBFDO-%lu", DeviceNumber);
	RtlInitUnicodeString(&DeviceName, DeviceBuffer);

	Status = IoCreateDevice(DriverObject,
				sizeof(OHCI_DEVICE_EXTENSION),
            /* + DriverExtension->InitializationData.HwDeviceExtensionSize*/
				&DeviceName,
				FILE_DEVICE_CONTROLLER,
				0,
				FALSE,
				&fdo);

	if (!NT_SUCCESS(Status))
	{
		DPRINT1("IoCreateDevice call failed with status 0x%08x\n", Status);
		return Status;
	}

	// zerofill device extension
	DeviceExtension = (POHCI_DEVICE_EXTENSION)fdo->DeviceExtension;
	RtlZeroMemory(DeviceExtension, sizeof(OHCI_DEVICE_EXTENSION));

	/* Register device interface */
	Status = IoRegisterDeviceInterface(
		pdo,
		&GUID_DEVINTERFACE_USB_HOST_CONTROLLER,
		NULL,
		&DeviceExtension->HcdInterfaceName);
	if (!NT_SUCCESS(Status))
	{
		DPRINT("UHCI: IoRegisterDeviceInterface() failed with status 0x%08lx\n", Status);
		IoDeleteDevice(fdo);
		return Status;
	}

	DeviceExtension->NextDeviceObject = IoAttachDeviceToDeviceStack(fdo, pdo);
       DPRINT1("UHCI: PDO %p [%wZ]\n",pdo, &pdo->DriverObject->DriverName);
       DPRINT1("UHCI: FDO %p [%wZ]\n",fdo, &fdo->DriverObject->DriverName);

       DPRINT1("UHCI: Lower device %p [%wZ]\n",DeviceExtension->NextDeviceObject,
							&DeviceExtension->NextDeviceObject->DriverObject->DriverName);
	fdo->Flags &= ~DO_DEVICE_INITIALIZING;

	// Initialize device extension
	DeviceExtension->IsFDO = TRUE;
	DeviceExtension->DeviceNumber = DeviceNumber;
	DeviceExtension->PhysicalDeviceObject = pdo;
	DeviceExtension->FunctionalDeviceObject = fdo;
	DeviceExtension->DriverExtension = DriverExtension;

//	swprintf(LinkDeviceBuffer, L"\\DosDevice\\HCD%lu", DeviceNumber++);
	swprintf(LinkDeviceBuffer, L"\\??\\HCD%lu", DeviceNumber++);

	RtlInitUnicodeString(&LinkDeviceName, LinkDeviceBuffer);

	Status = IoCreateSymbolicLink(&LinkDeviceName, &DeviceName);

	if (!NT_SUCCESS(Status))
	{
		DPRINT1("IoCreateSymbolicLink call failed with status 0x%08x\n", Status);
		return Status;
	}

	DPRINT1("Done AddDevice\n");
	return STATUS_SUCCESS;
}

VOID STDCALL 
DriverUnload(PDRIVER_OBJECT DriverObject)
{
	POHCI_DEVICE_EXTENSION DeviceExtension;
	PDEVICE_OBJECT DeviceObject;
	struct pci_dev *dev;	

	DeviceObject = DriverObject->DeviceObject;
	DeviceExtension = (POHCI_DEVICE_EXTENSION)DeviceObject->DeviceExtension;

	dev = DeviceExtension->pdev;

	DPRINT1("DriverUnload()\n");

	// Exit usb device
	usb_exit();

	// Remove device (ohci_pci_driver.remove)
	uhci_pci_driver.remove(dev);

	ExFreePool(dev->slot_name);
	ExFreePool(dev);

	// Perform some cleanup
	uhci_hcd_cleanup();
}

NTSTATUS
InitLinuxWrapper(PDEVICE_OBJECT DeviceObject)
{
	NTSTATUS Status = STATUS_SUCCESS;

	POHCI_DEVICE_EXTENSION DeviceExtension = (POHCI_DEVICE_EXTENSION)DeviceObject->DeviceExtension;
	struct pci_dev *dev;

	// Fill generic linux structs
	dev = ExAllocatePoolWithTag(PagedPool, sizeof(struct pci_dev), USB_UHCI_TAG);

	DeviceExtension->pdev = dev;

//	init_wrapper(dev);
	dev->irq = DeviceExtension->InterruptVector;

	dev->dev_ext = (PVOID)DeviceExtension;

	dev->slot_name = ExAllocatePoolWithTag(NonPagedPool, 128, USB_UHCI_TAG); // 128 max len for slot name

	strcpy(dev->dev.name, "UnivHCI PCI-USB Controller");
	strcpy(dev->slot_name, "UHCD PCI Slot");

	// Init the OHCI HCD. Probe will be called automatically, but will fail because id=NULL
//	Status = uhci_hcd_init();
	//FIXME: Check status returned value, Need to convert Unix errno to Win.
	// some thing like -> LINERR2NTSTATUS(x);

	// Init core usb
//	usb_init();

	// Probe device with real id now
//	uhci_pci_driver.probe(dev, uhci_pci_ids);

//	DPRINT1("SysIoBusNumA %d\n",DeviceExtension->SystemIoBusNumber);

//	DeviceExtension->SystemIoBusNumber = dev->bus->number;

//	DPRINT1("SysIoBusNumB %d\n",DeviceExtension->SystemIoBusNumber);

	DPRINT1("InitLinuxWrapper() done\n");

	return Status; 
}	

NTSTATUS STDCALL
OHCD_PnPStartDevice(IN PDEVICE_OBJECT DeviceObject,
					IN PIRP	Irp)
{
	PIO_STACK_LOCATION Stack = IoGetCurrentIrpStackLocation(Irp);
	PDRIVER_OBJECT DriverObject;
	POHCI_DRIVER_EXTENSION DriverExtension;
	POHCI_DEVICE_EXTENSION DeviceExtension;
	PCM_RESOURCE_LIST AllocatedResources;

	/*
	* Get the initialization data we saved in VideoPortInitialize.
	*/
	DriverObject = DeviceObject->DriverObject;
	DriverExtension	= IoGetDriverObjectExtension(DriverObject, DriverObject);
	DeviceExtension	= (POHCI_DEVICE_EXTENSION)DeviceObject->DeviceExtension;

	/*
	* Store	some resources in the DeviceExtension.
	*/
	AllocatedResources = Stack->Parameters.StartDevice.AllocatedResources;
	if (AllocatedResources != NULL)
	{
		CM_FULL_RESOURCE_DESCRIPTOR	*FullList;
		CM_PARTIAL_RESOURCE_DESCRIPTOR *Descriptor;
		ULONG ResourceCount;
		ULONG ResourceListSize;

		/* Save the resource list */
		ResourceCount = AllocatedResources->List[0].PartialResourceList.Count;

		ResourceListSize = FIELD_OFFSET(CM_RESOURCE_LIST, List[0].PartialResourceList.
			PartialDescriptors[ResourceCount]);

		DeviceExtension->AllocatedResources = ExAllocatePool(PagedPool, ResourceListSize);
		if (DeviceExtension->AllocatedResources == NULL)
		{
			return STATUS_INSUFFICIENT_RESOURCES;
		}

		RtlCopyMemory(DeviceExtension->AllocatedResources,
			AllocatedResources,
			ResourceListSize);

		/* Get the interrupt level/vector -	needed by HwFindAdapter	sometimes */
		for	(FullList =	AllocatedResources->List;
			FullList < AllocatedResources->List	+ AllocatedResources->Count;
			FullList++)
		{
			/* FIXME: Is this ASSERT ok	for	resources from the PNP manager?	*/
			/*ASSERT(FullList->InterfaceType == PCIBus &&
				FullList->BusNumber	== DeviceExtension->SystemIoBusNumber &&
				1 == FullList->PartialResourceList.Version &&
				1 == FullList->PartialResourceList.Revision);*/
			DPRINT1("AllocRess->Count: %d, PartResList.Count: %d\n",
					AllocatedResources->Count, FullList->PartialResourceList.Count);

			for	(Descriptor	= FullList->PartialResourceList.PartialDescriptors;
				Descriptor < FullList->PartialResourceList.PartialDescriptors + FullList->PartialResourceList.Count;
				Descriptor++)
			{
				if (Descriptor->Type ==	CmResourceTypeInterrupt)
				{
					DeviceExtension->InterruptLevel	= Descriptor->u.Interrupt.Level;
					DeviceExtension->InterruptVector = Descriptor->u.Interrupt.Vector;

	                DPRINT1("Interrupt level: 0x%x Interrupt	Vector:	0x%x\n",
                                 DeviceExtension->InterruptLevel,
                                 DeviceExtension->InterruptVector);
				}
				else if (Descriptor->Type == CmResourceTypePort)
				{
					DeviceExtension->BaseAddress	= Descriptor->u.Port.Start;
					DeviceExtension->BaseAddrLength = Descriptor->u.Port.Length;
					DeviceExtension->Flags          = Descriptor->Flags;

					((struct hc_driver *)uhci_pci_ids->driver_data)->flags &= ~HCD_MEMORY;
					
     				DPRINT1("I/O resource: start=0x%x, length=0x%x\n",
                                 DeviceExtension->BaseAddress.u.LowPart, DeviceExtension->BaseAddrLength);
				}
				else if (Descriptor->Type == CmResourceTypeMemory)
				{
					DeviceExtension->BaseAddress	= Descriptor->u.Memory.Start;
					DeviceExtension->BaseAddrLength = Descriptor->u.Memory.Length;
					DeviceExtension->Flags          = Descriptor->Flags;

					((struct hc_driver *)uhci_pci_ids->driver_data)->flags |= HCD_MEMORY;
					
     				DPRINT1("Memory resource: start=0x%x, length=0x%x\n",
                                 DeviceExtension->BaseAddress.u.LowPart, DeviceExtension->BaseAddrLength);
				}
				else
     				DPRINT1("Get resource type: %d, Generic start=0x%x Generic length=0x%x\n",
	     				Descriptor->Type, Descriptor->u.Generic.Start.u.LowPart, Descriptor->u.Generic.Length);
				
			}
		}
	}

	/*
	* Init wrapper with this object
	*/
	return InitLinuxWrapper(DeviceObject);
}
NTSTATUS STDCALL
ForwardIrpAndWaitCompletion(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp,
	IN PVOID Context)
{
	if (Irp->PendingReturned)
		KeSetEvent((PKEVENT)Context, IO_NO_INCREMENT, FALSE);
	return STATUS_MORE_PROCESSING_REQUIRED;
}

NTSTATUS
ForwardIrpAndWait(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp)
{
	PDEVICE_OBJECT LowerDevice = ((POHCI_DEVICE_EXTENSION)DeviceObject->DeviceExtension)->NextDeviceObject;
	KEVENT Event;
	NTSTATUS Status;

	ASSERT(LowerDevice);

	KeInitializeEvent(&Event, NotificationEvent, FALSE);
	IoCopyCurrentIrpStackLocationToNext(Irp);

	DPRINT1("UHCI: Calling lower device %p [%wZ]\n", LowerDevice, &LowerDevice->DriverObject->DriverName);
	IoSetCompletionRoutine(Irp, ForwardIrpAndWaitCompletion, &Event, TRUE, TRUE, TRUE);

	Status = IoCallDriver(LowerDevice, Irp);
	if (Status == STATUS_PENDING)
	{
		Status = KeWaitForSingleObject(&Event, Suspended, KernelMode, FALSE, NULL);
		if (NT_SUCCESS(Status))
			Status = Irp->IoStatus.Status;
	}

	return Status;
}

NTSTATUS STDCALL
ForwardIrpAndForget(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp)
{
	PDEVICE_OBJECT LowerDevice = ((POHCI_DEVICE_EXTENSION)DeviceObject->DeviceExtension)->NextDeviceObject;

	ASSERT(LowerDevice);

	IoSkipCurrentIrpStackLocation(Irp);
	return IoCallDriver(LowerDevice, Irp);
}

// Dispatch PNP
NTSTATUS STDCALL
DispatchPnp(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
   PIO_STACK_LOCATION IrpSp;
   NTSTATUS Status;
   ULONG MinorFunction;
   ULONG_PTR Information = 0;

   IrpSp = IoGetCurrentIrpStackLocation(Irp);
   MinorFunction = IrpSp->MinorFunction;

    switch (MinorFunction)
    {
     case IRP_MN_START_DEVICE:
          {
         Status = ForwardIrpAndWait(DeviceObject, Irp);
         if (NT_SUCCESS(Status) && NT_SUCCESS(Irp->IoStatus.Status))
         	 Status = OHCD_PnPStartDevice(DeviceObject, Irp);
         Irp->IoStatus.Status = Status;
         Irp->IoStatus.Information = 0;
         IoCompleteRequest(Irp, IO_NO_INCREMENT);
         break;
       }

      case IRP_MN_REMOVE_DEVICE:
      case IRP_MN_QUERY_REMOVE_DEVICE:
      case IRP_MN_CANCEL_REMOVE_DEVICE:
      case IRP_MN_SURPRISE_REMOVAL:

      case IRP_MN_STOP_DEVICE:
      {
         Status = ForwardIrpAndWait(DeviceObject, Irp);
         if (NT_SUCCESS(Status) && NT_SUCCESS(Irp->IoStatus.Status))
            Status = STATUS_SUCCESS;
         Irp->IoStatus.Status = Status;
         Irp->IoStatus.Information = 0;
         IoCompleteRequest(Irp, IO_NO_INCREMENT);

		 IoDeleteDevice(DeviceObject); // just delete device for now
         break;
        }
      case IRP_MN_QUERY_STOP_DEVICE:
      case IRP_MN_CANCEL_STOP_DEVICE:
        {
         Status = STATUS_SUCCESS;
         Irp->IoStatus.Status = STATUS_SUCCESS;
         Irp->IoStatus.Information = 0;
         IoCompleteRequest(Irp, IO_NO_INCREMENT);
         break;
         }
	case IRP_MN_QUERY_DEVICE_RELATIONS: /* (optional) 0x7 */
	{
		switch (IrpSp->Parameters.QueryDeviceRelations.Type)
		{
			case BusRelations:
			{
				DPRINT("UHCI: IRP_MJ_PNP / IRP_MN_QUERY_DEVICE_RELATIONS / BusRelations\n");
				return ForwardIrpAndForget(DeviceObject, Irp);
			}
			case RemovalRelations:
			{
				DPRINT("UHCI: IRP_MJ_PNP / IRP_MN_QUERY_DEVICE_RELATIONS / RemovalRelations\n");
				return ForwardIrpAndForget(DeviceObject, Irp);
			}
			default:
				DPRINT1("UHCI: IRP_MJ_PNP / IRP_MN_QUERY_DEVICE_RELATIONS / Unknown type 0x%lx\n",
					IrpSp->Parameters.QueryDeviceRelations.Type);
				return ForwardIrpAndForget(DeviceObject, Irp);
		}
		break;
	}

      default:
         {
	DPRINT1("UHCI: unknown minor function 0x%x\n", MinorFunction);
	return ForwardIrpAndForget(DeviceObject, Irp);
              Status = STATUS_NOT_IMPLEMENTED;
              break;
          }
   }
   	Irp->IoStatus.Information = Information;
	Irp->IoStatus.Status = Status;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);
	return Status;
}

NTSTATUS STDCALL 
DispatchPower(PDEVICE_OBJECT fido, PIRP Irp)
{
	DbgPrint("IRP_MJ_POWER dispatch\n");
        Irp->IoStatus.Information = 0;
        Irp->IoStatus.Status = STATUS_SUCCESS;
        IoCompleteRequest(Irp, IO_NO_INCREMENT);
	return STATUS_SUCCESS;
}

NTSTATUS STDCALL
UhciCleanup(
        IN PDEVICE_OBJECT DeviceObject,
        IN PIRP Irp)
{
        DPRINT1("UHCI: IRP_MJ_CLEANUP\n");
        Irp->IoStatus.Information = 0;
        Irp->IoStatus.Status = STATUS_SUCCESS;
        IoCompleteRequest(Irp, IO_NO_INCREMENT);
        return STATUS_SUCCESS;
}

NTSTATUS STDCALL
IrpStub(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp)
{
	NTSTATUS Status = STATUS_NOT_SUPPORTED;

	if (((POHCI_DEVICE_EXTENSION)DeviceObject->DeviceExtension)->IsFDO)
	{
		/* Forward some IRPs to lower device */
		switch (IoGetCurrentIrpStackLocation(Irp)->MajorFunction)
		{
			case IRP_MJ_CREATE:
			{
				Status = UhciCreate(DeviceObject, Irp);
				break;
			}
			case IRP_MJ_CLOSE:
			{
				Status = UhciClose(DeviceObject, Irp);
				break;
			}
			case IRP_MJ_CLEANUP:
			{
				Status = UhciCleanup(DeviceObject, Irp);
				break;
			}
			case IRP_MJ_DEVICE_CONTROL:
			{
				Status = UsbDeviceControl(DeviceObject, Irp);
				break;
			}
			case IRP_MJ_READ:
			case IRP_MJ_WRITE:
				DPRINT1("UHCI: FDO stub for major function 0x%lx\n",
					IoGetCurrentIrpStackLocation(Irp)->MajorFunction);
				return ForwardIrpAndForget(DeviceObject, Irp);
			default:
			{
				DPRINT1("UHCI: FDO stub for major function 0x%lx\n",
					IoGetCurrentIrpStackLocation(Irp)->MajorFunction);
#ifndef NDEBUG
				DbgBreakPoint();
#endif
				Status = Irp->IoStatus.Status;
			}
		}
	}
	else
	{
		/* Forward some IRPs to attached FDO */
		switch (IoGetCurrentIrpStackLocation(Irp)->MajorFunction)
		{
			case IRP_MJ_CREATE:
			case IRP_MJ_CLOSE:
			case IRP_MJ_CLEANUP:
			case IRP_MJ_READ:
			case IRP_MJ_WRITE:
			case IRP_MJ_DEVICE_CONTROL:
				DPRINT1("UHCI: PDO stub for major function 0x%lx\n",
					IoGetCurrentIrpStackLocation(Irp)->MajorFunction);
				return ForwardIrpAndForget(DeviceObject, Irp);
			default:
			{
				DPRINT1("UHCI: PDO stub for major function 0x%lx\n",
					IoGetCurrentIrpStackLocation(Irp)->MajorFunction);
#ifndef NDEBUG
				DbgBreakPoint();
#endif
				Status = Irp->IoStatus.Status;
			}
		}
	}

	Irp->IoStatus.Status = Status;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);
	return Status;
}
/*
 * Standard DriverEntry method.
 */
NTSTATUS STDCALL
DriverEntry(IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegPath)
{
   ULONG i;
   DPRINT1("******************** Cromwell UHCI ********************\n");

   DriverObject->DriverUnload = DriverUnload;
   DriverObject->DriverExtension->AddDevice = AddDevice;

	for (i = 0; i < IRP_MJ_MAXIMUM_FUNCTION; i++)
		DriverObject->MajorFunction[i] = IrpStub;

   DriverObject->MajorFunction[IRP_MJ_PNP] = DispatchPnp;
   DriverObject->MajorFunction[IRP_MJ_POWER] = DispatchPower;

   return STATUS_SUCCESS;
}
